"""Top-level package for ape-apeman."""

from .main import wrangler
from .version import __author__, __email__, __timestamp__, __version__

__all__ = ["wrangler", __version__, __timestamp__, __author__, __email__]



# cli for ape-apeman

import sys

import click
from box import Box
from eth_utils import from_wei, to_checksum_address

from .context import APE
from .exception_handler import ExceptionHandler
from .json import dumps
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"


def output(ctx, data):
    if ctx.obj.json:
        kwargs = dict(humanize=ctx.obj.humanize, hex_bytes=True)
        if ctx.obj.compact:
            kwargs["separators"] = (",", ":")
        else:
            kwargs["indent"] = 2
        data = dumps(data, **kwargs)
    click.echo(data)


def fail(message):
    click.echo(f"Failed: {message}", err=True)
    sys.exit(1)


@click.group(name="apeman")
@click.version_option(message=header)
@click.option(
    "-e",
    "--ecosystem",
    type=str,
    required=True,
    envvar="APE_ECOSYSTEM",
    show_envvar=True,
    default="ethereum",
)
@click.option(
    "-n",
    "--network",
    type=str,
    required=True,
    envvar="APE_NETWORK",
    show_envvar=True,
)
@click.option(
    "-p",
    "--provider",
    type=str,
    required=True,
    envvar="APE_PROVIDER",
    show_envvar=True,
)
@click.option("-d", "--debug", is_flag=True, help="debug mode")
@click.option(
    "-j/-r",
    "--json/--raw",
    is_flag=True,
    default=True,
    help="select output format",
)
@click.option("-c", "--compact", is_flag=True, help="output compact json")
@click.option(
    "-h", "--humanize", is_flag=True, help="human-friendly (but lossy) output"
)
@click.pass_context
def cli(ctx, ecosystem, network, debug, provider, json, compact, humanize):
    ctx.obj = Box(ehandler=ExceptionHandler(debug))
    ctx.obj.debug = debug
    ctx.obj.json = json
    ctx.obj.compact = compact
    ctx.obj.humanize = humanize
    ctx.obj.ape = APE(ecosystem, network, provider)


@cli.command
@click.option("-u", "--url", is_flag=True, help="output transaction url")
@click.option(
    "-l", "--logs", is_flag=True, help="output decoded transaction logs"
)
@click.argument("txn-hash", type=str)
@click.pass_context
def txn(ctx, url, logs, txn_hash):
    """output transaction receipt"""
    with ctx.obj.ape as ape:
        if url:
            ret = ape.explorer.get_transaction_url(txn_hash)
        elif logs:
            receipt = ape.provider.get_receipt(txn_hash)
            logs = receipt.decode_logs()
            ret = [log.dict() for log in logs]
        else:
            receipt = ape.provider.get_receipt(txn_hash)
            ret = receipt.dict()

    output(ctx, ret)


@cli.command
@click.argument("account", type=str)
@click.option("-w", "--wei", is_flag=True, help="output wei")
@click.option("-g", "--gwei", is_flag=True, help="output gwei")
@click.option("-e", "--ether", is_flag=True, help="output ether")
@click.pass_context
def balance(ctx, account, wei, gwei, ether):
    """output account balance"""
    account = to_checksum_address(account)
    with ctx.obj.ape as ape:
        as_wei = ape.provider.get_balance(account)
    as_ether = from_wei(as_wei, "ether")
    as_gwei = from_wei(as_wei, "gwei")
    if wei:
        ret = as_wei
    elif gwei:
        ret = as_gwei
    elif ether:
        ret = as_ether
    else:
        ret = {account: dict(wei=as_wei, gwei=as_gwei, ether=as_ether)}

    output(ctx, ret)


@cli.group
@click.pass_context
def eth(ctx):
    """expose web3.eth methods"""
    pass


@eth.command
@click.pass_context
def get_block_number(ctx):
    with ctx.obj.ape as ape:
        ret = ape.web3.eth.get_block_number()
    output(ctx, ret)


@eth.command
@click.argument("block", type=str)
@click.pass_context
def get_block(ctx, block):
    if block.isnumeric():
        block = int(block)
    with ctx.obj.ape as ape:
        _block = ape.web3.eth.get_block(block)
        ret = dict(_block)
    output(ctx, ret)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import logging
import os
from pathlib import Path

import pytest


@pytest.fixture(autouse=True)
def setup_logger():
    logging.basicConfig(
        format="%(levelname)s %(name)s.%(funcName)s: %(message)s [%(filename)s:%(lineno)s]",
        force=True,
    )
    logging.getLogger("urllib3.connectionpool").setLevel("WARNING")
    logging.getLogger("ape").setLevel("ERROR")


@pytest.fixture(autouse=True)
def prevent_home_data_dir():
    home_ape = Path.home().resolve() / ".ape"
    cwd_ape = Path(".").resolve() / "ape"
    assert not home_ape.exists()
    assert not cwd_ape.exists()
    try:
        yield True
    finally:
        assert not home_ape.exists()
        assert not cwd_ape.exists()


@pytest.fixture
def txn_hash():
    return "0x96823521c0c1999e4c1023279a1b4ef343649d37ac5f5b50a5fc7264064cb0ac"


@pytest.fixture
def contract_address():
    return "0xbBeb16aB5F13cd52bDa5d7AD49be4E7677231680"


@pytest.fixture
def owner_address():
    return "0x27566e752e56D403fB436C8b7031e7fcc75b6b4f"


@pytest.fixture
def owner_private_key():
    return os.environ["TEST_ACCOUNT_PRIVATE_KEY"]



# test use from python code

import logging
from pathlib import Path
from pprint import pformat
from subprocess import run

import pytest
from ape.exceptions import ContractLogicError
from ape_ethereum.transactions import Receipt
from eth_account import Account
from eth_utils import is_same_address

import ape_apeman.context

info = logging.info


@pytest.fixture
def APE(monkeypatch, shared_datadir):
    monkeypatch.setenv("APE_PROJECT_DIR", str(shared_datadir / "ape_project"))
    monkeypatch.setenv("APE_DATA_DIR", str(shared_datadir / "ape_data"))
    return ape_apeman.context.APE


@pytest.fixture
def check_url():
    def _check_url(url):
        assert isinstance(url, str)
        assert url.startswith("https://")

    return _check_url


@pytest.fixture
def check_receipt():
    def _check_receipt(receipt):
        assert isinstance(receipt, Receipt)

    return _check_receipt


@pytest.fixture
def check_block_number():
    def _check_block_number(block_number):
        assert isinstance(block_number, int)
        assert block_number > 1

    return _check_block_number


def test_module_context_url(APE, txn_hash, check_url):
    with APE() as ape:
        url = ape.explorer.get_transaction_url(txn_hash)
    check_url(url)


def test_module_init_url(APE, txn_hash, check_url):
    ape = APE(connect=True)
    url = ape.explorer.get_transaction_url(txn_hash)
    check_url(url)


def test_module_connect_url(APE, txn_hash, check_url):
    ape = APE()
    ape.connect()
    url = ape.explorer.get_transaction_url(txn_hash)
    check_url(url)


def test_module_context_receipt(APE, txn_hash, check_receipt):
    with APE() as ape:
        receipt = ape.provider.get_receipt(txn_hash)
    check_receipt(receipt)


def test_module_init_receipt(APE, txn_hash, check_receipt):
    ape = APE(connect=True)
    receipt = ape.provider.get_receipt(txn_hash)
    check_receipt(receipt)


def test_module_connect_receipt(APE, txn_hash, check_receipt):
    ape = APE()
    ape.connect()
    receipt = ape.provider.get_receipt(txn_hash)
    check_receipt(receipt)


def test_module_context_web3(APE, check_block_number):
    with APE() as ape:
        n = ape.web3.eth.get_block_number()
    check_block_number(n)


def test_module_init_web3(APE, check_block_number):
    ape = APE(connect=True)
    n = ape.web3.eth.get_block_number()
    check_block_number(n)


def test_module_connect_web3(APE, check_block_number):
    ape = APE()
    ape.connect()
    n = ape.web3.eth.get_block_number()
    check_block_number(n)


def test_module_context_contract(APE, contract_address):
    with APE() as ape:
        contract = ape.contracts.instance_at(contract_address)
        assert contract
        symbol = contract.symbol()
        assert symbol == "ETHERSIEVE"
        print(f"{contract_address=} {symbol=}")


def test_module_init_contract(APE, contract_address):
    ape = APE(connect=True)
    contract = ape.contracts.instance_at(contract_address)
    assert contract
    symbol = contract.symbol()
    assert symbol == "ETHERSIEVE"
    print(f"{contract_address=} {symbol=}")


def test_module_connect_contract(APE, contract_address):
    ape = APE()
    ape.connect()
    contract = ape.contracts.instance_at(contract_address)
    assert contract
    symbol = contract.symbol()
    assert symbol == "ETHERSIEVE"
    print(f"{contract_address=} {symbol=}")


def _sign_and_send(ape, contract_address, owner_address, owner_private_key):
    account = Account().from_key(owner_private_key)
    assert account.address == owner_address
    contract = ape.Contract(contract_address)
    nonce_before = ape.provider.get_nonce(account.address)
    txn = contract.getPrices.as_transaction(
        sender=account.address,
        required_confirmations=1,
        max_priority_fee="10 gwei",
    )
    txn.nonce = ape.provider.get_nonce(account.address)
    txn.signature = account.sign_transaction(txn.dict())
    receipt = ape.provider.send_transaction(txn)
    print(receipt)
    ret = receipt.return_value
    print(ret)
    nonce_after = ape.provider.get_nonce(account.address)
    assert nonce_before != nonce_after


@pytest.mark.uses_gas
def test_module_context_sign_and_send(
    APE, contract_address, owner_address, owner_private_key
):
    with APE() as ape:
        _sign_and_send(ape, contract_address, owner_address, owner_private_key)


@pytest.mark.uses_gas
def test_module_init_sign_and_send(
    APE, contract_address, owner_address, owner_private_key
):
    ape = APE(connect=True)
    _sign_and_send(ape, contract_address, owner_address, owner_private_key)


@pytest.mark.uses_gas
def test_module_connect_sign_and_send(
    APE, contract_address, owner_address, owner_private_key
):
    ape = APE()
    ape.connect()
    _sign_and_send(ape, contract_address, owner_address, owner_private_key)


def pdebug(object):
    logging.debug(pformat(object))


def _find_files(dir, label=None, output=pdebug):
    dir = Path(dir).resolve()
    label = label or str(dir)
    proc = run(
        ["find", str(dir), "-type", "f"],
        check=True,
        text=True,
        capture_output=True,
    )
    assert not proc.stderr
    files = proc.stdout.strip().split("\n")
    if output:
        output({label: files})
    return files


def test_module_account_call(
    APE, shared_datadir, contract_address, owner_address, owner_private_key
):
    with APE() as ape:
        # call a public function

        contract = ape.Contract(contract_address)
        symbol = contract.symbol()
        assert symbol == "ETHERSIEVE"

        # call a function requiring authorized caller address
        with pytest.raises(ContractLogicError) as exc:
            components = contract.getComponents()
        assert exc
        assert "is missing role" in str(exc.value)

        # specify the public address in the lookup
        components = contract.getComponents(sender=owner_address)
        assert components

        # create an ape account
        account = ape.account(private_key=owner_private_key)
        assert is_same_address(account.address, owner_address)

        # with account.autosign_enabled():
        components = contract.getComponents(sender=account.address)
        assert components

        prices = contract.getPrices(sender=account.address)
        assert prices

        with account:
            prices = contract.getPrices(sender=account.address)
        assert prices

        with account.autosign_enabled():
            prices = contract.getPrices(sender=account.address)
        assert prices

        info(pformat(components))
        info(pformat(prices))



# ape_apeman contract module


class ContractCallResult:
    def __init__(self, ret=None):
        self.decoded_logs = []
        self.receipt = None
        self.ret = ret

    def set_receipt(self, receipt):
        self.receipt = receipt
        self.decoded_logs = [log for log in receipt.decode_logs()]
        if len(self.decoded_logs):
            self.ret = self.decoded_logs[0]
        else:
            self.ret = None

    def dict(self):
        return dict(
            ret=self.ret, receipt=self.receipt.dict() if self.receipt else None
        )



# stateless ape account

import atexit
import json
from pathlib import Path
from secrets import token_hex

from eth_account import Account
from eth_utils import to_normalized_address


class KeyAccount:
    def __init__(
        self,
        *,
        ape,
        private_key,
        alias=None,
        password=None,
        autosign=False,
    ):
        self.ape = ape
        self.autosign = autosign
        self.alias = alias or token_hex(16)
        self.password = password or token_hex(32)
        account = Account.encrypt(private_key, self.password)
        self.address = to_normalized_address(account["address"])
        accounts_dir = self.ape.accounts.containers["accounts"].data_folder
        self.keyfile = Path(accounts_dir) / f"{self.alias}.json"
        self.keyfile.write_text(json.dumps(account))
        atexit.register(self.keyfile.unlink, missing_ok=True)
        self.ape_account = None
        self.load()

    def __del__(self):
        if self.keyfile:
            self.keyfile.unlink(missing_ok=True)

    def autosign_enabled(self):
        self.autosign = True
        return self

    def __enter__(self):
        return self.load()

    def load(self, autosign=False):
        if self.ape_account is None:
            self.ape_account = self.ape.accounts.load(self.alias)
        if self.autosign or autosign:
            self.ape_account.set_autosign(True, self.password)
        return self.ape_account

    def __exit__(self, *args, **kwargs):
        self.ape_account.set_autosign(False)
        self.autosign = False



# contract call tests

import logging
from pprint import pformat

import pytest
from eth_utils import to_wei

import ape_apeman
from ape_apeman.context import ContractCallResult

info = logging.info

TIP_GWEI = 10


@pytest.fixture
def ape():
    with ape_apeman.wrangler() as ape:
        yield ape


def test_call_lookup(ape, contract_address):
    contract = ape.Contract(contract_address)
    symbol = contract.symbol()
    assert symbol == "ETHERSIEVE"


@pytest.mark.tryfirst
def test_call_set(ape, contract_address, owner_address, owner_private_key):
    royalty_fee = 1000000000000000
    result = ape.call_contract(
        contract_address,
        "setPrintRoyalty",
        royalty_fee,
        private_key=owner_private_key,
        max_priority_fee=ape.provider.priority_fee + to_wei(TIP_GWEI, "gwei"),
    )
    assert isinstance(result, ContractCallResult)
    info(pformat(result.ret))
    info(pformat(result.receipt))


def test_call_get_set(ape, contract_address, owner_address, owner_private_key):
    # get the current prices
    pre_prices = ape.call_contract(
        contract_address, "getPrices", sender=owner_address
    )
    info(f"{pre_prices=}")
    assert pre_prices.ret.printRoyalty == 1000000000000000
    assert pre_prices.receipt is None

    # set the royalty price
    ret_set = ape.call_contract(
        contract_address,
        "setPrintRoyalty",
        13337,
        sender=owner_address,
        private_key=owner_private_key,
        autosign=True,
        max_priority_fee=ape.provider.priority_fee + to_wei(TIP_GWEI, "gwei"),
    )
    assert ret_set.ret is None
    assert ret_set.receipt

    # get the changed price
    tmp_prices = ape.call_contract(
        contract_address, "getPrices", sender=owner_address
    )
    info(f"{tmp_prices=}")
    assert tmp_prices.ret.printRoyalty == 13337
    assert tmp_prices.receipt is None

    # reset the price back to the original value
    ret_reset = ape.call_contract(
        contract_address,
        "setPrintRoyalty",
        pre_prices.ret.printRoyalty,
        private_key=owner_private_key,
        autosign=True,
        max_priority_fee=ape.provider.priority_fee + to_wei(TIP_GWEI, "gwei"),
    )
    assert ret_reset.ret is None
    assert ret_reset.receipt

    # finally test that the price is restored to original value
    post_prices = ape.call_contract(
        contract_address, "getPrices", sender=owner_address
    )
    info(f"{post_prices=}")
    assert post_prices.ret.printRoyalty == pre_prices.ret.printRoyalty
    assert post_prices.receipt is None


def test_call_mint(ape, contract_address, owner_address, owner_private_key):
    uri = "ipfs://bafkreie5npehcwbdbog3j4w7otw22jzm6szsjtihtjnhc5oxhvwukhczgm"
    result = ape.call_contract(
        contract_address,
        "safeMint",
        owner_address,
        uri,
        private_key=owner_private_key,
        max_priority_fee=ape.provider.priority_fee + to_wei(TIP_GWEI, "gwei"),
    )
    assert isinstance(result, ContractCallResult)
    info(pformat(result.dict()))



# ape context manager

import atexit
import logging
import os
import shutil
import tempfile
from pathlib import Path

import ape
from ape.contracts.base import ContractCallHandler, ContractTransactionHandler
from eth_utils import is_same_address

from .account import KeyAccount
from .contract import ContractCallResult

logger = logging.getLogger(__name__)
logger.setLevel(os.environ.get("APEMAN_LOG_LEVEL", "WARNING"))


class APE:
    def __init__(
        self,
        ecosystem=None,
        network=None,
        provider=None,
        selector=None,
        connect=False,
        project_dir=None,
        data_dir=None,
    ):
        self.connection = None
        self.project_dir = self.init_dir(project_dir, "APE_PROJECT_DIR")
        self.data_dir = self.init_dir(data_dir, "APE_DATA_DIR")

        ape.config.DATA_FOLDER = self.data_dir
        ape.config.PROJECT_FOLDER = self.project_dir

        ape.project = ape.Project(self.project_dir)

        ape.config.load(force_reload=True)
        if selector is None:
            ecosystem = ecosystem or os.environ["APE_ECOSYSTEM"]
            network = network or os.environ["APE_NETWORK"]
            provider = provider or os.environ["APE_PROVIDER"]
            selector = os.environ.get(
                "APE_SELECTOR", f"{ecosystem}:{network}:{provider}"
            )
        self.context_manager = ape.networks.parse_network_choice(selector)
        if connect is True:
            self.connect()

    def init_dir(self, param, key):
        dir = param or os.environ.get(key)
        if dir:
            dir = Path(dir)
            dir.mkdir(exist_ok=True)
        else:
            dir = Path(tempfile.mkdtemp()).resolve()
            atexit.register(shutil.rmtree, str(dir))
            logger.debug(f"created temp directory {str(dir)} as {key}")

        if isinstance(dir, Path) is False or dir.is_dir() is False:
            raise TypeError(f"{dir} is not a directory")
        dir.chmod(0o700)
        return dir

    def __del__(self):
        if self.connection:
            self.disconnect()

    def account(
        self,
        private_key,
        alias=None,
        password=None,
        autosign=False,
    ):
        return KeyAccount(
            ape=self,
            private_key=private_key,
            alias=alias,
            password=password,
            autosign=autosign,
        )

    def connect(self, *args, **kwargs):
        if self.connection is None:

            self.connection = self.context_manager.__enter__(*args, **kwargs)

            self.__all__ = ape.__all__

            for attr in ape.__all__:
                setattr(self, attr, getattr(ape, attr))

            self.provider = self.connection
            self.network = self.connection.network
            self.explorer = self.provider.network.explorer
            self.web3 = self.provider.web3
            self.contracts = self.network.chain_manager.contracts

            # assert self.provider is self.project.provider
            # assert self.network is self.project.provider.network
            # assert self.explorer is self.project.provider.network.explorer
            # assert self.web3 is self.project.provider.web3
            # assert self.contracts is self.project.provider.network.chain_manager.contracts

            logger.debug(f"connected: {self}")

        return self

    def disconnect(self, *args, **kwargs):
        if self.connection:
            self.context_manager.__exit__(*args, **kwargs)
            logger.debug(f"disconnected: {self}")
        self.connection = None

    def __enter__(self, *args, **kwargs):
        return self.connect(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        self.disconnect(*args, **kwargs)

    def call_contract(self, contract_address, function_name, *args, **kwargs):
        """call a lookup or mutable contract function, returning a ContractCallResult

        kwargs:
          sender (ADDRESS if lookup, PRIVATE_KEY if mutable/transaction
          autosign: (bool) automatically sign transaction
          max_fee:  max fee per GAS to be used
          max_priority_fee: specify amount of 'tip' per GAS
        """
        result = ContractCallResult()
        contract = self.Contract(contract_address)
        function = getattr(contract, function_name)
        if isinstance(function, ContractCallHandler):
            # lookup call may include sender=ADDRESS
            result.ret = function(*args, **kwargs)
        elif isinstance(function, ContractTransactionHandler):
            # mutable call uses transaction, sender must be a private_key
            sender = kwargs.pop("sender", None)
            private_key = kwargs.pop("private_key", None)
            account = kwargs.pop("account", None)
            if private_key:
                account = self.account(private_key=private_key)

            elif account:
                account = account
            else:
                raise RuntimeError(
                    "mutable call requires 'private_key' or 'account'"
                )

            account.autosign = kwargs.pop("autosign", True)

            if sender is not None:
                if not is_same_address(sender, account.address):
                    raise ValueError(
                        f"account key mismatches sender address {sender}"
                    )

            with account as ape_account:
                receipt = function(*args, **kwargs, sender=ape_account)
                result.set_receipt(receipt)

        else:
            raise TypeError(
                f"function {function_name} not found in contract at {contract_address}"
            )

        return result
